# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library used by example_viz.py to generate visualizations.

This file illustrates the following:
  - How to subclass an existing agent to add visualization functionality.
    - For DQN we visualize the cumulative rewards and the Q-values for each
      action (MyDQNAgent).
    - For Rainbow we visualize the cumulative rewards and the Q-value
      distributions for each action (MyRainbowAgent).
  - How to subclass Runner to run in eval mode, lay out the different subplots,
    generate the visualizations, and compile them into a video (MyRunner).
  - The function `run()` is the main entrypoint for running everything.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import logging

from dopamine.agents.dqn import dqn_agent
from dopamine.agents.rainbow import rainbow_agent
from dopamine.discrete_domains import atari_lib
from dopamine.discrete_domains import iteration_statistics
from dopamine.discrete_domains import run_experiment
from dopamine.utils import agent_visualizer
from dopamine.utils import atari_plotter
from dopamine.utils import bar_plotter
from dopamine.utils import line_plotter
import gin
import numpy as np
import tensorflow as tf
import tf_slim


class MyDQNAgent(dqn_agent.DQNAgent):
  """Sample DQN agent to visualize Q-values and rewards."""

  def __init__(self, sess, num_actions, summary_writer=None):
    super(MyDQNAgent, self).__init__(sess, num_actions,
                                     summary_writer=summary_writer)
    self.q_values = [[] for _ in range(num_actions)]
    self.rewards = []

  def step(self, reward, observation):
    self.rewards.append(reward)
    return super(MyDQNAgent, self).step(reward, observation)

  def _select_action(self):
    action = super(MyDQNAgent, self)._select_action()
    q_vals = self._sess.run(self._net_outputs.q_values,
                            {self.state_ph: self.state})[0]
    for i in range(len(q_vals)):
      self.q_values[i].append(q_vals[i])
    return action

  def reload_checkpoint(self, checkpoint_path, use_legacy_checkpoint=False):
    if use_legacy_checkpoint:
      variables_to_restore = atari_lib.maybe_transform_variable_names(
          tf.compat.v1.global_variables(), legacy_checkpoint_load=True)
    else:
      global_vars = set([x.name for x in tf.compat.v1.global_variables()])
      ckpt_vars = [
          '{}:0'.format(name)
          for name, _ in tf.train.list_variables(checkpoint_path)
      ]
      include_vars = list(global_vars.intersection(set(ckpt_vars)))
      variables_to_restore = tf_slim.get_variables_to_restore(
          include=include_vars)
    if variables_to_restore:
      reloader = tf.compat.v1.train.Saver(var_list=variables_to_restore)
      reloader.restore(self._sess, checkpoint_path)
      logging.info('Done restoring from %s', checkpoint_path)
    else:
      logging.info('Nothing to restore!')

  def get_q_values(self):
    return self.q_values

  def get_rewards(self):
    return [np.cumsum(self.rewards)]


class MyRainbowAgent(rainbow_agent.RainbowAgent):
  """Sample Rainbow agent to visualize Q-values and rewards."""

  def __init__(self, sess, num_actions, summary_writer=None):
    super(MyRainbowAgent, self).__init__(sess, num_actions,
                                         summary_writer=summary_writer)
    self.rewards = []

  def step(self, reward, observation):
    self.rewards.append(reward)
    return super(MyRainbowAgent, self).step(reward, observation)

  def reload_checkpoint(self, checkpoint_path, use_legacy_checkpoint=False):
    if use_legacy_checkpoint:
      variables_to_restore = atari_lib.maybe_transform_variable_names(
          tf.compat.v1.global_variables(), legacy_checkpoint_load=True)
    else:
      global_vars = set([x.name for x in tf.compat.v1.global_variables()])
      ckpt_vars = [
          '{}:0'.format(name)
          for name, _ in tf.train.list_variables(checkpoint_path)
      ]
      include_vars = list(global_vars.intersection(set(ckpt_vars)))
      variables_to_restore = tf_slim.get_variables_to_restore(
          include=include_vars)
    if variables_to_restore:
      reloader = tf.compat.v1.train.Saver(var_list=variables_to_restore)
      reloader.restore(self._sess, checkpoint_path)
      logging.info('Done restoring from %s', checkpoint_path)
    else:
      logging.info('Nothing to restore!')

  def get_probabilities(self):
    return self._sess.run(tf.squeeze(self._net_outputs.probabilities),
                          {self.state_ph: self.state})

  def get_rewards(self):
    return [np.cumsum(self.rewards)]


class MyRunner(run_experiment.Runner):
  """Sample Runner class to generate visualizations."""

  def __init__(self, base_dir, trained_agent_ckpt_path, create_agent_fn,
               use_legacy_checkpoint=False):
    self._trained_agent_ckpt_path = trained_agent_ckpt_path
    self._use_legacy_checkpoint = use_legacy_checkpoint
    super(MyRunner, self).__init__(base_dir, create_agent_fn)

  def _initialize_checkpointer_and_maybe_resume(self, checkpoint_file_prefix):
    self._agent.reload_checkpoint(self._trained_agent_ckpt_path,
                                  self._use_legacy_checkpoint)
    self._start_iteration = 0

  def _run_one_iteration(self, iteration):
    statistics = iteration_statistics.IterationStatistics()
    logging.info('Starting iteration %d', iteration)
    _, _ = self._run_eval_phase(statistics)
    return statistics.data_lists

  def visualize(self, record_path, num_global_steps=500):
    if not tf.io.gfile.exists(record_path):
      tf.io.gfile.makedirs(record_path)
    self._agent.eval_mode = True

    # Set up the game playback rendering.
    atari_params = {'environment': self._environment}
    atari_plot = atari_plotter.AtariPlotter(parameter_dict=atari_params)
    # Plot the rewards received next to it.
    reward_params = {'x': atari_plot.parameters['width'],
                     'xlabel': 'Timestep',
                     'ylabel': 'Reward',
                     'title': 'Rewards',
                     'get_line_data_fn': self._agent.get_rewards}
    reward_plot = line_plotter.LinePlotter(parameter_dict=reward_params)
    action_names = [
        'Action {}'.format(x) for x in range(self._agent.num_actions)]
    # Plot Q-values (DQN) or Q-value distributions (Rainbow).
    q_params = {'x': atari_plot.parameters['width'] // 2,
                'y': atari_plot.parameters['height'],
                'legend': action_names}
    if 'DQN' in self._agent.__class__.__name__:
      q_params['xlabel'] = 'Timestep'
      q_params['ylabel'] = 'Q-Value'
      q_params['title'] = 'Q-Values'
      q_params['get_line_data_fn'] = self._agent.get_q_values
      q_plot = line_plotter.LinePlotter(parameter_dict=q_params)
    elif 'Implicit' in self._agent.__class__.__name__:
      q_params['xlabel'] = 'Timestep'
      q_params['ylabel'] = 'Quantile Value'
      q_params['title'] = 'Quantile Values'
      q_params['get_line_data_fn'] = self._agent.get_q_values
      q_plot = line_plotter.LinePlotter(parameter_dict=q_params)
    else:
      q_params['xlabel'] = 'Return'
      q_params['ylabel'] = 'Return probability'
      q_params['title'] = 'Return distribution'
      q_params['get_bar_data_fn'] = self._agent.get_probabilities
      q_plot = bar_plotter.BarPlotter(parameter_dict=q_params)
    screen_width = (
        atari_plot.parameters['width'] + reward_plot.parameters['width'])
    screen_height = (
        atari_plot.parameters['height'] + q_plot.parameters['height'])
    # Dimensions need to be divisible by 2:
    if screen_width % 2 > 0:
      screen_width += 1
    if screen_height % 2 > 0:
      screen_height += 1
    visualizer = agent_visualizer.AgentVisualizer(
        record_path=record_path, plotters=[atari_plot, reward_plot, q_plot],
        screen_width=screen_width, screen_height=screen_height)
    global_step = 0
    while global_step < num_global_steps:
      initial_observation = self._environment.reset()
      action = self._agent.begin_episode(initial_observation)
      while True:
        observation, reward, is_terminal, _ = self._environment.step(action)
        global_step += 1
        visualizer.visualize()
        if self._environment.game_over or global_step >= num_global_steps:
          break
        elif is_terminal:
          self._agent.end_episode(reward)
          action = self._agent.begin_episode(observation)
        else:
          action = self._agent.step(reward, observation)
      self._end_episode(reward)
    visualizer.generate_video()


def create_dqn_agent(sess, environment, summary_writer=None):
  return MyDQNAgent(sess, num_actions=environment.action_space.n,
                    summary_writer=summary_writer)


def create_rainbow_agent(sess, environment, summary_writer=None):
  return MyRainbowAgent(sess, num_actions=environment.action_space.n,
                        summary_writer=summary_writer)


def create_runner(base_dir, trained_agent_ckpt_path, agent='dqn',
                  use_legacy_checkpoint=False):
  create_agent = create_dqn_agent if agent == 'dqn' else create_rainbow_agent
  return MyRunner(base_dir, trained_agent_ckpt_path, create_agent,
                  use_legacy_checkpoint)


def run(agent, game, num_steps, root_dir, restore_ckpt,
        use_legacy_checkpoint=False):
  """Main entrypoint for running and generating visualizations.

  Args:
    agent: str, agent type to use.
    game: str, Atari 2600 game to run.
    num_steps: int, number of steps to play game.
    root_dir: str, root directory where files will be stored.
    restore_ckpt: str, path to the checkpoint to reload.
    use_legacy_checkpoint: bool, whether to restore from a legacy (pre-Keras)
      checkpoint.
  """
  tf.compat.v1.reset_default_graph()
  config = """
  atari_lib.create_atari_environment.game_name = '{}'
  WrappedReplayBuffer.replay_capacity = 300
  """.format(game)
  base_dir = os.path.join(root_dir, 'agent_viz', game, agent)
  gin.parse_config(config)
  runner = create_runner(base_dir, restore_ckpt, agent, use_legacy_checkpoint)
  runner.visualize(os.path.join(base_dir, 'images'), num_global_steps=num_steps)



# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Setup script for Dopamine.

This script will install Dopamine as a Python module.

See: https://github.com/google/dopamine
"""

import pathlib
from setuptools import find_packages
from setuptools import setup

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

install_requires = [
    'tensorflow >= 2.2.0',
    'gin-config >= 0.3.0',
    'absl-py >= 0.9.0',
    'opencv-python >= 3.4.8.29',
    'gym[atari] >= 0.13.1',
    'flax >= 0.2.0',
    'jax >= 0.1.72',
    'jaxlib >= 0.1.51',
    'Pillow >= 7.0.0',
    'numpy >= 1.16.4',
    'pygame >= 1.9.2',
    'pandas >= 0.24.2',
    'tf_slim >= 1.0',
]

dopamine_description = (
    'Dopamine: A framework for flexible Reinforcement Learning research')

setup(
    name='dopamine_rl',
    version='3.1.8',
    description=dopamine_description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/google/dopamine',
    author='The Dopamine Team',
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',

        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',

    ],
    keywords='dopamine, reinforcement, machine, learning, research',
    include_package_data=True,
    packages=find_packages(exclude=['docs']),
    package_data={'testdata': ['testdata/*.gin']},
    install_requires=install_requires,
    project_urls={  # Optional
        'Documentation': 'https://github.com/google/dopamine',
        'Bug Reports': 'https://github.com/google/dopamine/issues',
        'Source': 'https://github.com/google/dopamine',
    },
    license='Apache 2.0',
)



# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An extension of Rainbow to perform quantile regression.

This loss is computed as in "Distributional Reinforcement Learning with Quantile
Regression" - Dabney et. al, 2017"
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools



from dopamine.jax import networks
from dopamine.jax.agents.dqn import dqn_agent
from dopamine.replay_memory import prioritized_replay_buffer
from flax import nn
import gin
import jax
import jax.numpy as jnp
import tensorflow as tf


@functools.partial(jax.vmap, in_axes=(None, 0, 0, 0, None))
def target_distribution(target_network, next_states, rewards, terminals,
                        cumulative_gamma):
  """Builds the Quantile target distribution as per Dabney et al. (2017).

  Args:
    target_network: Jax Module used for the target network.
    next_states: numpy array of batched next states.
    rewards: numpy array of batched rewards.
    terminals: numpy array of batched terminals.
    cumulative_gamma: float, cumulative gamma to use (static_argnum).

  Returns:
    The target distribution from the replay.
  """
  is_terminal_multiplier = 1. - terminals.astype(jnp.float32)
  # Incorporate terminal state to discount factor.
  gamma_with_terminal = cumulative_gamma * is_terminal_multiplier
  next_state_target_outputs = target_network(next_states)
  q_values = jnp.squeeze(next_state_target_outputs.q_values)
  next_qt_argmax = jnp.argmax(q_values)
  logits = jnp.squeeze(next_state_target_outputs.logits)
  next_logits = logits[next_qt_argmax]
  return jax.lax.stop_gradient(rewards + gamma_with_terminal * next_logits)


@functools.partial(jax.jit, static_argnums=(7, 8, 9))
def train(target_network, optimizer, states, actions, next_states, rewards,
          terminals, kappa, num_atoms, cumulative_gamma):
  """Run a training step."""
  def loss_fn(model, target, mean_loss=True):
    logits = jax.vmap(model)(states).logits
    logits = jnp.squeeze(logits)
    # Fetch the logits for its selected action. We use vmap to perform this
    # indexing across the batch.
    chosen_action_logits = jax.vmap(lambda x, y: x[y])(logits, actions)
    bellman_errors = (target[:, None, :] -
                      chosen_action_logits[:, :, None])  # Input `u' of Eq. 9.
    # Eq. 9 of paper.
    huber_loss = (
        (jnp.abs(bellman_errors) <= kappa).astype(jnp.float32) *
        0.5 * bellman_errors ** 2 +
        (jnp.abs(bellman_errors) > kappa).astype(jnp.float32) *
        kappa * (jnp.abs(bellman_errors) - 0.5 * kappa))

    tau_hat = ((jnp.arange(num_atoms, dtype=jnp.float32) + 0.5) /
               num_atoms)  # Quantile midpoints.  See Lemma 2 of paper.
    # Eq. 10 of paper.
    tau_bellman_diff = jnp.abs(
        tau_hat[None, :, None] - (bellman_errors < 0).astype(jnp.float32))
    quantile_huber_loss = tau_bellman_diff * huber_loss
    # Sum over tau dimension, average over target value dimension.
    loss = jnp.sum(jnp.mean(quantile_huber_loss, 2), 1)
    if mean_loss:
      loss = jnp.mean(loss)
    return loss
  grad_fn = jax.value_and_grad(loss_fn)
  target = target_distribution(target_network,
                               next_states,
                               rewards,
                               terminals,
                               cumulative_gamma)
  mean_loss, grad = grad_fn(optimizer.target, target)
  # Get the loss without taking its mean.
  loss = loss_fn(optimizer.target, target, mean_loss=False)
  optimizer = optimizer.apply_gradient(grad)
  return optimizer, loss, mean_loss


@gin.configurable
class JaxQuantileAgent(dqn_agent.JaxDQNAgent):
  """An implementation of Quantile regression DQN agent."""

  def __init__(self,
               num_actions,
               observation_shape=dqn_agent.NATURE_DQN_OBSERVATION_SHAPE,
               observation_dtype=dqn_agent.NATURE_DQN_DTYPE,
               stack_size=dqn_agent.NATURE_DQN_STACK_SIZE,
               network=networks.QuantileNetwork,
               kappa=1.0,
               num_atoms=200,
               gamma=0.99,
               update_horizon=1,
               min_replay_history=50000,
               update_period=4,
               target_update_period=10000,
               epsilon_fn=dqn_agent.linearly_decaying_epsilon,
               epsilon_train=0.1,
               epsilon_eval=0.05,
               epsilon_decay_period=1000000,
               replay_scheme='prioritized',
               optimizer='adam',
               summary_writer=None,
               summary_writing_frequency=500,
               allow_partial_reload=False):
    """Initializes the agent and constructs the Graph.

    Args:
      num_actions: Int, number of actions the agent can take at any state.
      observation_shape: tuple of ints or an int. If single int, the observation
        is assumed to be a 2D square.
      observation_dtype: DType, specifies the type of the observations. Note
        that if your inputs are continuous, you should set this to jnp.float32.
      stack_size: int, number of frames to use in state stack.
      network: tf.Keras.Model, expects 3 parameters: num_actions, num_atoms,
        network_type. A call to this object will return an instantiation of the
        network provided. The network returned can be run with different inputs
        to create different outputs. See
        dopamine.discrete_domains.jax.networks.QuantileNetwork as an example.
      kappa: Float, Huber loss cutoff.
      num_atoms: Int, the number of buckets for the value function distribution.
      gamma: Float, exponential decay factor as commonly used in the RL
        literature.
      update_horizon: Int, horizon at which updates are performed, the 'n' in
        n-step update.
      min_replay_history: Int, number of stored transitions for training to
        start.
      update_period: Int, period between DQN updates.
      target_update_period: Int, ppdate period for the target network.
      epsilon_fn: Function expecting 4 parameters: (decay_period, step,
        warmup_steps, epsilon), and which returns the epsilon value used for
        exploration during training.
      epsilon_train: Float, final epsilon for training.
      epsilon_eval: Float, epsilon during evaluation.
      epsilon_decay_period: Int, number of steps for epsilon to decay.
      replay_scheme: String, replay memory scheme to be used. Choices are:
        uniform - Standard (DQN) replay buffer (Mnih et al., 2015)
        prioritized - Prioritized replay buffer (Schaul et al., 2015)
      optimizer: str, name of optimizer to use.
      summary_writer: SummaryWriter object for outputting training statistics.
        Summary writing disabled if set to None.
      summary_writing_frequency: int, frequency with which summaries will be
        written. Lower values will result in slower training.
      allow_partial_reload: bool, whether we allow reloading a partial agent
        (for instance, only the network parameters).
    """
    self._num_atoms = num_atoms
    self._kappa = kappa
    self._replay_scheme = replay_scheme

    super(JaxQuantileAgent, self).__init__(
        num_actions=num_actions,
        observation_shape=observation_shape,
        observation_dtype=observation_dtype,
        stack_size=stack_size,
        network=network.partial(num_atoms=num_atoms),
        gamma=gamma,
        update_horizon=update_horizon,
        min_replay_history=min_replay_history,
        update_period=update_period,
        target_update_period=target_update_period,
        epsilon_fn=epsilon_fn,
        epsilon_train=epsilon_train,
        epsilon_eval=epsilon_eval,
        epsilon_decay_period=epsilon_decay_period,
        optimizer=optimizer,
        summary_writer=summary_writer,
        summary_writing_frequency=summary_writing_frequency,
        allow_partial_reload=allow_partial_reload)

  def _create_network(self, name):
    r"""Builds a Quantile ConvNet.

    Equivalent to Rainbow ConvNet, only now the output logits are interpreted
    as quantiles.

    Args:
      name: str, this name is passed to the Jax Module.
    Returns:
      network: Jax Model, the network instantiated by Jax.
    """
    _, initial_params = self.network.init(self._rng,
                                          name=name,
                                          x=self.state,
                                          num_actions=self.num_actions,
                                          num_atoms=self._num_atoms)
    return nn.Model(self.network, initial_params)

  def _build_replay_buffer(self):
    """Creates the replay buffer used by the agent."""
    if self._replay_scheme not in ['uniform', 'prioritized']:
      raise ValueError('Invalid replay scheme: {}'.format(self._replay_scheme))
    # Both replay schemes use the same data structure, but the 'uniform' scheme
    # sets all priorities to the same value (which yields uniform sampling).
    return prioritized_replay_buffer.OutOfGraphPrioritizedReplayBuffer(
        observation_shape=self.observation_shape,
        stack_size=self.stack_size,
        update_horizon=self.update_horizon,
        gamma=self.gamma,
        observation_dtype=self.observation_dtype)

  def _train_step(self):
    """Runs a single training step.

    Runs training if both:
      (1) A minimum number of frames have been added to the replay buffer.
      (2) `training_steps` is a multiple of `update_period`.

    Also, syncs weights from online_network to target_network if training steps
    is a multiple of target update period.
    """
    if self._replay.add_count > self.min_replay_history:
      if self.training_steps % self.update_period == 0:
        self._sample_from_replay_buffer()
        self.optimizer, loss, mean_loss = train(
            self.target_network,
            self.optimizer,
            self.replay_elements['state'],
            self.replay_elements['action'],
            self.replay_elements['next_state'],
            self.replay_elements['reward'],
            self.replay_elements['terminal'],
            self._kappa,
            self._num_atoms,
            self.cumulative_gamma)
        if self._replay_scheme == 'prioritized':
          # The original prioritized experience replay uses a linear exponent
          # schedule 0.4 -> 1.0. Comparing the schedule to a fixed exponent of
          # 0.5 on 5 games (Asterix, Pong, Q*Bert, Seaquest, Space Invaders)
          # suggested a fixed exponent actually performs better, except on Pong.
          probs = self.replay_elements['sampling_probabilities']
          loss_weights = 1.0 / jnp.sqrt(probs + 1e-10)
          loss_weights /= jnp.max(loss_weights)

          # Rainbow and prioritized replay are parametrized by an exponent
          # alpha, but in both cases it is set to 0.5 - for simplicity's sake we
          # leave it as is here, using the more direct sqrt(). Taking the square
          # root "makes sense", as we are dealing with a squared loss.  Add a
          # small nonzero value to the loss to avoid 0 priority items. While
          # technically this may be okay, setting all items to 0 priority will
          # cause troubles, and also result in 1.0 / 0.0 = NaN correction terms.
          self._replay.set_priority(self.replay_elements['indices'],
                                    jnp.sqrt(loss + 1e-10))

          # Weight the loss by the inverse priorities.
          loss = loss_weights * loss
          mean_loss = jnp.mean(loss)
        if self.summary_writer is not None:
          summary = tf.compat.v1.Summary(value=[
              tf.compat.v1.Summary.Value(tag='QuantileLoss',
                                         simple_value=mean_loss)])
          self.summary_writer.add_summary(summary, self.training_steps)
      if self.training_steps % self.target_update_period == 0:
        self._sync_weights()

    self.training_steps += 1

  def _store_transition(self,
                        last_observation,
                        action,
                        reward,
                        is_terminal,
                        priority=None):
    """Stores a transition when in training mode.

    Stores the following tuple in the replay buffer (last_observation, action,
    reward, is_terminal, priority).

    Args:
      last_observation: Last observation, type determined via observation_type
        parameter in the replay_memory constructor.
      action: An integer, the action taken.
      reward: A float, the reward.
      is_terminal: Boolean indicating if the current state is a terminal state.
      priority: Float. Priority of sampling the transition. If None, the default
        priority will be used. If replay scheme is uniform, the default priority
        is 1. If the replay scheme is prioritized, the default priority is the
        maximum ever seen [Schaul et al., 2015].
    """
    if priority is None:
      if self._replay_scheme == 'uniform':
        priority = 1.
      else:
        priority = self._replay.sum_tree.max_recorded_priority

    if not self.eval_mode:
      self._replay.add(last_observation, action, reward, is_terminal, priority)



# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""The entry point for running a Dopamine agent.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags
from absl import logging

from dopamine.discrete_domains import run_experiment
import tensorflow as tf


flags.DEFINE_string('base_dir', None,
                    'Base directory to host all required sub-directories.')
flags.DEFINE_multi_string(
    'gin_files', [], 'List of paths to gin configuration files (e.g.'
    '"dopamine/agents/dqn/dqn.gin").')
flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files '
    '(e.g. "DQNAgent.epsilon_train=0.1",'
    '      "create_environment.game_name="Pong"").')

FLAGS = flags.FLAGS




def main(unused_argv):
  """Main method.

  Args:
    unused_argv: Arguments (unused).
  """
  logging.set_verbosity(logging.INFO)
  tf.compat.v1.disable_v2_behavior()

  run_experiment.load_gin_configs(FLAGS.gin_files, FLAGS.gin_bindings)
  runner = run_experiment.create_runner(FLAGS.base_dir)
  runner.run_experiment()


if __name__ == '__main__':
  flags.mark_flag_as_required('base_dir')
  app.run(main)
